/*
 * FILE: $Source: /u/chestnut/ezk/kitot/advanced-os/2/RCS/execute.c,v $
 * HEADER: $Id: execute.c,v 1.13 1992/02/11 14:22:33 ezk Exp ezk $
 * REVISION HISTORY:
 * $Log: execute.c,v $
 * Revision 1.13  1992/02/11  14:22:33  ezk
 * one small fix to variable assignments.
 *
 * Revision 1.12  1992/02/11  13:56:39  ezk
 * FINAL VERSION (unless I get an extension...)
 *
 * Revision 1.11  1992/02/10  03:52:41  ezk
 * bg, fg, ^C, ^Z, jobs all work with multiple processes in job queue.
 *
 * Revision 1.10  1992/02/10  00:30:08  ezk
 * bg and & work now.  bug fixes so that status reported in "jobs" cmd
 * is correct, and also each time you hit newline.
 *
 * Revision 1.9  1992/02/09  19:01:19  ezk
 * more work on job control. can bg and stop process. some status gets
 * reported but not perfect yet.
 *
 * Revision 1.8  1992/02/09  00:22:13  ezk
 * one process in fg.  can ^C it or wait until it ends.  shell prompt
 * using PS1 if used.  can suspend a process but not bg/fg it yet.
 *
 * Revision 1.7  1992/02/08  08:04:04  ezk
 * very basic foreground only processes running.
 *
 * Revision 1.6  1992/02/08  04:30:01  ezk
 * basic pathname lookup done
 *
 * Revision 1.5  1992/02/08  02:02:10  ezk
 * most parsing done.  set/unset/export work.  variable assignment work.
 *
 * Revision 1.4  1992/02/07  05:24:16  ezk
 * unset command fully working
 *
 * Revision 1.3  1992/02/07  04:53:50  ezk
 * built major jobs structures and converted code to use it.
 * basic export command.
 *
 * Revision 1.2  1992/02/06  21:50:46  ezk
 * simple variables (w/o braces) are supported and expanded, even if
 * prefeixed or postfixed by other strings or variables.
 *
 * Revision 1.1  1992/02/06  02:55:43  ezk
 * Initial revision
 *
 * 
 * Execute processes and do job control with/on them.
 * 
 * E6118y - Advanced Operating Systems.
 * Erez Zadok.
 * Columbia U.  Spring 1992.
 */


#include "process.h"

				/* GLOBALS */
int cur_job, max_jobs, next_fg, next_bg;
jobs_t jobs[MAX_JOBS];

				/* LOCALS */
static enum sh_builtins_e sh_builtins;


				/* some local fxns */
void do_exit(), do_error(), do_default(), do_jobs(), do_export(), do_exec();
void do_set(), do_unset(), do_fg(), do_bg(), do_kill();

#ifdef ESH_DEBUG

void do_shtrace();
				/* print a pipeline */
void print_pipeline(cp)
     proc_t *cp;
{ proc_t *tcp;
  int count = 1;

  if (cp == NULL_PROC) {
    fprintf(stderr, "NULL_PROC!\n\n");
    return;
  }

  tcp = cp;

  while (tcp != NULL_PROC) {
    fprintf(stderr, "\t<%d> %sEXECUTE \"%s\" with ", count++,
	    CUR_BG(tcp) ? "BACKGROUND " : "", tcp->argv[0]);
    if (tcp->argc <= 1)
      fprintf(stderr, "no arguments.\n");
    else {
      register int i;
      fprintf(stderr, "argument(s) ");
      for (i=1; i<tcp->argc; ++i)
	fprintf(stderr, "\"%s\" ", tcp->argv[i]);
      fprintf(stderr, ".\n");
    }
    tcp = tcp->next;
    if (tcp != NULL_PROC)
      fprintf(stderr, "\tPIPELINED TO:\n");
  }				/* end of while loop */
/*  if (cp)
    fprintf(stderr, "\n");	/* pretty output */
}

				/* Turn parser and lexer tracing */ 
static void do_shtrace(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{ char *errmsg, *a2, *a3;
  short int stat;

  errmsg = "Usage: shtrace <all | lex | yacc> <on | off>";
  a2 = hcp->argv[1];
  a3 = hcp->argv[2];
  if (hcp && hcp->argc != 3  ||	/* don't allow things like "ls|exit|more" */
      strcmp(a2, "all") && strcmp(a2, "lex") && 
      strcmp(a2, "yacc")  ||  strcmp(a3, "on") && strcmp(a3, "off")) {
    do_error(hcp, icp, errmsg);
    return;
  }
				/* the next statement is ugly, but... */
  if (!strcmp(a2, "all")) {
    if (!strcmp(a3, "on"))
      yydebug = yy_flex_debug = 1;
    else if (!strcmp(a3, "off"))
      yydebug = yy_flex_debug = 0;
  } else if (!strcmp(a2, "lex")) {
    if (!strcmp(a3, "on"))
      yy_flex_debug = 1;
    else if (!strcmp(a3, "off"))
      yy_flex_debug = 0;
  } else if (!strcmp(a2, "yacc")) {
    if (!strcmp(a3, "on"))
      yydebug = 1;
    else if (!strcmp(a3, "off"))
      yydebug = 0;
  }
}

#endif /* ESH_DEBUG */

				/* execute a pipeline */
void exec_pipeline(cpp)
     proc_t **cpp;
{ proc_t *tcp;

#ifdef ESH_DEBUG
  print_pipeline(*cpp);		/* for debugging and tracing */
#endif /* ESH_DEBUG */

  if (!cpp)			/* sanity */
    return;
		/* actually execute commands, builtins first */
  switch (get_builtin_token((*cpp)->argv[0])) {
  case CMD_EXIT:
    do_exit(*cpp, *cpp);	/* will exit only if typed twice */
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;

#ifdef ESH_DEBUG
  case CMD_SHTRACE:
    do_shtrace(*cpp, *cpp);
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;
#endif /* ESH_DEBUG */

  case CMD_FG:
    do_fg(*cpp, *cpp);
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;

  case CMD_BG:
    do_bg(*cpp, *cpp);
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;

  case CMD_KILL:
    do_kill(*cpp, *cpp);
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;

  case CMD_JOBS:
    do_jobs(*cpp, *cpp);
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;

  case CMD_EXPORT:
    do_export(*cpp, *cpp);
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;

  case CMD_EXEC:
    do_exec(*cpp, *cpp);
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;

  case CMD_SET:
    do_set(*cpp, *cpp);
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;

  case CMD_UNSET:
    do_unset(*cpp, *cpp);
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;

  case CMD_ERROR:
    do_error(*cpp, *cpp, "null builtin cmd attempted execution(?!)");
    add_stat_to_proc(cpp, P_NOCHANGE, P_EXIT, 0);
    break;

  case CMD_NOTBUILTIN:		/* must be a real command to execute */
  default:
    do_default(*cpp, *cpp);
    break;			/* not needed but... */
  }
  each_line();			/* check of process status change */
  next_job_slot();		/* prepare for next job */
}

				/* terminate the shell and its sub-procs */
static void do_exit(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{
  if (hcp->next) {
    fprintf(stderr, "exit: makes no sense in a pipe\n");
    return;
  }  
  if (hcp && hcp->argc > 2) {	/* don't allow things like "ls|exit|more" */
    do_error(hcp, icp, "Usage: exit [return value]");
    return;
  }

  if (hcp->argc == 1) {		/* was kill given an argument */
    free_pipeline(jobs[cur_job].head_proc);	/* not really needed */
    exit(0);			/* good terminating return code */
  } else {
    int i = atoi(hcp->argv[1]);
    free_pipeline(jobs[cur_job].head_proc);	/* not really needed */
    exit(i);			/* return a selected error code */
  }
}

				/* bring a process to the foreground */
static void do_fg(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{ char *s;
  int i, status, retval, child_pid;
  proc_t *tcp;

  if (hcp->next) {
    fprintf(stderr, "fg: makes no sense in a pipe\n");
    return;
  }
  if ((hcp->argc != 1) &&
      (hcp->argc != 2  ||
       hcp->argc == 2  &&
       hcp->argv[1][0] != '%'  ||
       (i = to_number(&(hcp->argv[1][1]))) <= 0)) {
    do_error(hcp, icp, "Usage: fg [%job number]");
    return;
  }

  if (hcp->argc == 1) {		/* then fg the next_fg one */
    i = next_fg;
    next_fg = -1;
  } else {
    --i;			/* to get proper ref into job table */
  }
				/* find job's pid to background */
  if (jobs[i].head_proc == NULL_PROC  ||
      jobs[i].head_proc != NULL_PROC  &&  jobs[i].head_proc->argc == 2  &&
      strcmp(jobs[i].head_proc->argv[0], "fg") == 0) {
    do_error(hcp, icp, "no such job");
    return;
  }
  if (!CUR_STOP(jobs[i].head_proc) && !CUR_BG(jobs[i].head_proc)) {
    do_error(hcp, icp, "job is not stopped/backgrounded");
    return;
  }
  if (CUR_STOP(jobs[i].head_proc)) { /* then start process as well */
    status = kill(jobs[i].head_proc->pid, SIGCONT);
    if (status < 0) {			/* kill failed */
      PERROR("bg");
    }
				/* update status as running in fg */
    clear_proc_stat(jobs[i].head_proc);
  }

				/* take control over tty */
  signal(SIGTTOU, SIG_IGN);
  status = tcsetpgrp(open_fd(), jobs[i].head_proc->pid);
/*   status = tcsetpgrp(open_fd(), getpid()); */
  if (status < 0) {		/* cannot set terminal process group */
    PERROR("CHILD: tcsetpgrp failed");
    exit(errno);
  }
  status = -2;
  while (status != jobs[i].head_proc->pid) {
/* puts("BEFORE WAIT"); */
    signal(SIGINT, SIG_DFL);
    signal(SIGTSTP, SIG_DFL);
    status = wait3(&retval, WUNTRACED , NULL_RUSAGE);
    signal(SIGINT, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);
/* puts("AFTER WAIT"); */
    if (status <= 0)
      continue;
    tcp = get_head_proc(status);
				/* update status of processes */
/*      printf("EZK: %d %d %d\n",
	     tcp->cur_stat, tcp->new_stat, tcp->wait3_val);
*/
    if (WIFEXITED(retval)) {
      add_stat_to_proc(tcp, P_NOCHANGE, P_EXIT, retval);
    }
    if (WIFSTOPPED(retval)) {
      add_stat_to_proc(tcp, P_NOCHANGE, P_STOP, retval);
    }
    if (WIFSIGNALED(retval)) {
      add_stat_to_proc(tcp, P_NOCHANGE, P_SIG, retval);
    }
/*      printf("EZK: %d %d %d\n",
	     tcp->cur_stat, tcp->new_stat, tcp->wait3_val);
*/
#ifdef ESH_DEBUG
/*      fprintf(stderr, "PARENT(wait): status=%d\n", status);
      if (WIFEXITED(retval))
	fprintf(stderr, "PARENT(wait): PID %d exited with status %d\n",
		status, WEXITSTATUS(retval));
      if (WIFSTOPPED(retval))
	fprintf(stderr, "PARENT(wait): PID %d stopped by signal %d\n",
		status, WSTOPSIG(retval));
      if (WIFSIGNALED(retval))
	fprintf(stderr, "PARENT(wait): PID %d terminated by signal %d\n",
		status, WTERMSIG(retval));
*/
#endif /* ESH_DEBUG */
  }				/* end of while (status != child_pid)... */
  if (status < 0) {		/* wait ended due to an error */
    PERROR("PARENT: wait3 error");
    free_pipeline(tcp);
    return;
  }

/*
 * Get back control over terminal.
 */
  signal(SIGTTOU, SIG_IGN);
  status = tcsetpgrp(open_fd(), getpid());	/* pid is the shell's pid */
  if (status < 0) {		/* cannot set terminal process group */
    PERROR("PARENT: tcsetpgrp failed");
    return;
  }
				/* set process status as running */
/*   clear_proc_stat(jobs[i].head_proc); */
/*   jobs[i].head_proc->wait3_val = 0; */
/*   add_stat_to_proc(jobs[i].head_proc, P_NOCHANGE, P_BG, 0); */

/*   printf("EZK.2: %s %d %d %d\n", jobs[i].head_proc->argv[0],
	 jobs[i].head_proc->cur_stat, jobs[i].head_proc->new_stat,
	 jobs[i].head_proc->wait3_val); */

/*   signal(SIGINT, SIG_DFL); */
/*   signal(SIGTSTP, SIG_DFL); */
}

				/* put a stopped process in the background */
static void do_bg(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{ char *s;
  int i, status;
  proc_t *tcp;

  if (hcp->next) {
    fprintf(stderr, "bg: makes no sense in a pipe\n");
    return;
  }
  if ((hcp->argc != 1) &&
      (hcp->argc != 2  ||
       hcp->argc == 2  &&
       hcp->argv[1][0] != '%'  ||
       (i = to_number(&(hcp->argv[1][1]))) <= 0)) {
    do_error(hcp, icp, "Usage: bg [%job number]");
    return;
  }

  if (hcp->argc == 1) {		/* then bg the next_bg one */
     i = next_bg;
     next_bg = -1;
  } else {
    --i;			/* to get proper ref into job table */
  }
				/* find job's pid to background */
  if (jobs[i].head_proc == NULL_PROC  ||
      jobs[i].head_proc != NULL_PROC  &&  jobs[i].head_proc->argc == 2  &&
      strcmp(jobs[i].head_proc->argv[0], "bg") == 0) {
    do_error(hcp, icp, "no such job");
    return;
  }
  if (!CUR_STOP(jobs[i].head_proc)) {
    do_error(hcp, icp, "job is not stopped");
    return;
  }

  status = kill(jobs[i].head_proc->pid, SIGCONT);
  if (status < 0) {			/* kill failed */
    PERROR("bg");
  }
  next_fg = i;			/* so I know what to fg by default */

/*
 * Get back control over terminal.
 */
  signal(SIGTTOU, SIG_IGN);	/* pid is the shell's pid */
  status = tcsetpgrp(open_fd(), getpid());
  if (status < 0) {		/* cannot set terminal process group */
    PERROR("DO_BG: tcsetpgrp failed");
    return;
  }

				/* set process status as running */
/*   del_stat_from_proc(jobs[i].head_proc, P_BG, P_BG, 0); */
  clear_proc_stat(jobs[i].head_proc);
  jobs[i].head_proc->wait3_val = 0;
  add_stat_to_proc(jobs[i].head_proc, P_NOCHANGE, P_BG, 0);

/*   printf("EZK.2: %s %d %d %d\n", jobs[i].head_proc->argv[0],
	 jobs[i].head_proc->cur_stat, jobs[i].head_proc->new_stat,
	 jobs[i].head_proc->wait3_val); */
}

				/* kill a stopped or backgrounded process */
static void do_kill(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{
  if (hcp->next) {
    fprintf(stderr, "kill: makes no sense in a pipe\n");
    return;
  }
  puts("\t DOING KILL...");
}

				/* list jobs and their status */
static void do_jobs(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{ register int i;
  if (hcp->next) {
    fprintf(stderr, "jobs: makes no sense in a pipe\n");
    return;
  }
  for (i=0; i<max_jobs; ++i) {
    if (jobs[i].head_proc != NULL_PROC  &&
	(jobs[i].head_proc->argc != 1 ||
	 strcmp(jobs[i].head_proc->argv[0], "jobs")))
      print_proc_status(i, jobs[i].head_proc);
  }
}

				/* export a variable */
static void do_export(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{ register int i, a;
  char *s;

  if (hcp->next) {
    fprintf(stderr, "export: makes no sense in a pipe\n");
    return;
  }
  if (hcp->argc == 1) {		/* print list */
    printf("EXPORTED VARIABLES:\n");
    print_shenvp(sh_envp, shenvp_len);
    return;
  }

  for (a=1; a<hcp->argc; ++a) {
    s = hcp->argv[a];
				/* look for variable */
    i = bin_search_env(s, sh_envp, shenvp_len);
    if (i >= 0)			/* already exported.  do nothing. */
      continue;
				/* check unexported list */
    i = bin_search_env(s, unexported, unexp_len);
    if (i >= 0) {
      if (strstr(s, "=") != NULL_STR) /* then export whatever they gave us */
	add_var(sh_envp, &shenvp_len, s);
      else			/* export the old value */
	add_var(sh_envp, &shenvp_len, unexported[i]);
				/* remove old entry from unexported */
      del_var(unexported, &unexp_len, i);
    } else {			/* not found anywhere */
      add_var(sh_envp, &shenvp_len, s);
    }
  }				/* end of for loop */
}

				/* exec a new command */
static void do_exec(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{
  if (hcp->next) {
    fprintf(stderr, "exec: makes no sense in a pipe\n");
    return;
  }  

}

				/* print list of all variables */
static void do_set(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{
  if (hcp->next) {
    fprintf(stderr, "set: makes no sense in a pipe\n");
    return;
  }  
  if (hcp->argc == 1) {
    printf("EXPORTED VARIABLES:\n");
    print_shenvp(sh_envp, shenvp_len);
    printf("\nUNEXPORTED VARIABLES:\n");
    print_shenvp(unexported, unexp_len);
  } else {
    fprintf(stderr, "set: no support for extended command yet\n");
  }
}

				/* unset a variable */
static void do_unset(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{ register int i;

  if (hcp->next) {
    fprintf(stderr, "unset: makes no sense in a pipe\n");
    return;
  }  
  if (hcp->argc == 1) {
    printf("\nUNEXPORTED VARIABLES:\n");
    print_shenvp(unexported, unexp_len);
  }
  for (i=1; i<hcp->argc; ++i) {
    if (!strcmp(hcp->argv[i], "PATH") ||
	!strcmp(hcp->argv[i], "PS1") ||
	!strcmp(hcp->argv[i], "PS2") ||
	!strcmp(hcp->argv[i], "MAILCHECK") ||
	!strcmp(hcp->argv[i], "IFS")) {
      fprintf(stderr, "cannot unset %s\n", hcp->argv[i]);
      continue;
    } else {
      unset_var(hcp, icp, hcp->argv[i]);
    }
  }    
}

				/* print error message and cleanup */
static void do_error(hcp, icp, errmsg)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
     char *errmsg;
{
  fprintf(stderr, "%s: %s.\n", progname, errmsg);
  free_pipeline(jobs[cur_job].head_proc);	/* free mem */
  tmp_proc = jobs[cur_job].cur_proc = jobs[cur_job].head_proc = NULL_PROC;
}

				/* execute the command */
static void do_default(hcp, icp)
     proc_t *hcp, *icp;		/* head current proc and internal cur proc */
{ register int i, is_var;
  proc_t *tcp;
/*
 * This fxn has to decide when and if to update max_jobs and cur_job.
 * Later on I should make actual copy of envp for exec-ed procs.
 */

  /* what about hcp->next ... ? */
				/* check if we tried to set any variables */
  is_var = 0;
  for (i=0; i<hcp->argc; ++i) {
    if (var_set_cmd(hcp->argv[i])) {
      set_var(hcp->argv[i]);
      is_var = 1;
    }
  }
  if (is_var) return;
				/* see if this is a pathname to execute */
  tcp = hcp;
  while (tcp) {
    tcp->path = resolve_path(tcp->argv[0]);
    if (!tcp->path) { /* cleanup structures */
      fprintf(stderr, "%s: %s: not found\n", progname, tcp->argv[0]);
      free_pipeline(hcp);	/* if one path is bad, discard pipeline */
      return;
    }
    tcp = tcp->next;
  }
/*   setup_redirections(hcp, icp); */
/*   fprintf(stderr, "Found in \"%s\"\n", p); */
  if (hcp->next)		/* then a pipeline */
    start_pipeline(hcp, icp);
  else
    start_proc(hcp->path, hcp, icp);
}


/*
 * convert a string to a positive integer, return -1 if error.
 */
int to_number(s)
     char *s;
{ register int i;
  if (strlen(s) < 1)
    return -1;
  for (i=0; i<strlen(s); ++i)
    if (!isdigit(s[i]))
      return -1;
  return atoi(s);
}
