/*
 * FILE: $Source: /u/chestnut/ezk/kitot/advanced-os/2/RCS/job_ctrl.c,v $
 * HEADER: $Id: job_ctrl.c,v 1.8 1992/02/11 13:56:39 ezk Exp ezk $
 * REVISION HISTORY:
 * $Log: job_ctrl.c,v $
 * Revision 1.8  1992/02/11  13:56:39  ezk
 * FINAL VERSION (unless I get an extension...)
 *
 * Revision 1.7  1992/02/10  03:52:41  ezk
 * bg, fg, ^C, ^Z, jobs all work with multiple processes in job queue.
 *
 * Revision 1.6  1992/02/10  00:30:08  ezk
 * bg and & work now.  bug fixes so that status reported in "jobs" cmd
 * is correct, and also each time you hit newline.
 *
 * Revision 1.5  1992/02/09  19:01:19  ezk
 * more work on job control. can bg and stop process. some status gets
 * reported but not perfect yet.
 *
 * Revision 1.4  1992/02/09  00:22:13  ezk
 * one process in fg.  can ^C it or wait until it ends.  shell prompt
 * using PS1 if used.  can suspend a process but not bg/fg it yet.
 *
 * Revision 1.3  1992/02/08  08:04:04  ezk
 * very basic foreground only processes running.
 *
 * Revision 1.2  1992/02/08  04:30:01  ezk
 * basic pathname lookup done
 *
 * Revision 1.1  1992/02/08  02:08:41  ezk
 * Initial revision
 *
 *
 * 
 * Job control.
 * 
 * E6118y - Advanced Operating Systems.
 * Erez Zadok.
 * Columbia U.  Spring 1992.
 */


#include "process.h"

				/* GLOBALS */

				/* LOCALS */
static void status_check();




/*
 * Search the path for the p program.  NULL if not found.  Full path if
 * found.
 */
char *resolve_path(p)
     char *p;
{ static char *vn, s[MAXPATHLEN], *tmp;
  char *dot;
				/* If path contains '/' return it */
  if (strstr(p, "/"))
    return p;
				/* get PATH */
  strcpy(s, sh_getenv("PATH"));
  tmp = (char *) calloc(MAXPATHLEN, sizeof(char));
  if (!tmp) {
    fprintf(stderr, "no more memory for resolve_path\n");
    exit(-1);
  }  
				/* scan path for ':' delimiters */
  dot = strstr(s, "::");	/* strtok skips these and we need it */
  vn = strtok(s, ":");
  if (dot  &&  dot < vn) {
    strcpy(tmp, "./");
    strcat(tmp, p);
    if (!access(tmp, X_OK)) {
      return tmp;
    }
    dot = NULL;
  }
  while (vn != NULL) {
    if (dot  &&  dot < vn) {
      strcpy(tmp, "./");
      strcat(tmp, p);
      if (!access(tmp, X_OK)) {
	return tmp;
      }
      dot = NULL;
    }
    strcpy(tmp, vn);
    strcat(tmp, "/");
    strcat(tmp, p);
    if (!access(tmp, X_OK)) {
      return tmp;
    }
    vn = strtok(NULL, ":");
  }
  if (!vn && dot) {	/* last possibility is that :: is last in PATH */
    strcpy(tmp, "./");
    strcat(tmp, p);
    if (!access(tmp, X_OK)) {
      return tmp;
    }
    dot = NULL;
  }
  free(tmp);
  tmp = NULL_STR;
  return NULL_STR;
}


/*
 * Return any open fd.
 */
int open_fd()
{ register int i;
  for (i=0; i<MAX_FD; ++i)
    if (!feof(&_iob[i]) && !ferror(&_iob[i]))
      return i;
			/* else open /dev/tty and return fd if can */
  i = open("/dev/tty", O_RDONLY);
  if (i < 0) {
    PERROR("cannot open \"/dev/tty\"");
    exit(i);
  }
  return i;
}


/*
 * Start a process.
 */
void start_proc(path, hcp, icp)
     char *path;
     proc_t *hcp, *icp;
{ int child_pid, pid, ppid, pgrp, tcgrp, status, retval;
  proc_t *tcp;

  if (child_pid = fork()) {		/* in PARENT */
    pid = getpid();
    ppid = getppid();
    pgrp = getpgrp(pid);
    tcgrp = tcgetpgrp(open_fd());

#ifdef ESH_DEBUG
/*    fprintf(stderr, "PARENT: CHILD=%d PID=%d PPID=%d PGID=%d TCGRP=%d.\n",
	    child_pid, pid, ppid, pgrp, tcgrp);
*/
#endif /* ESH_DEBUG */
    if (child_pid < 0) {	/* bad fork for some reason */
      PERROR("fork failed");
      free_pipeline(hcp);
      return;
    }
/*
 * In order for the child to receive signals that do not affect the parent
 * shell, we must change the process group of the child first, before
 * execve-ing it.
 * For example, a SIGTERM (^C) should terminate the child process (if it has
 * not trapped that one) and the parent should be informed of that.
 */
    status = setpgrp(child_pid, child_pid);
    if (status < 0) {		/* cannot set the process group */
      PERROR("PARENT: setpgrp failed");
      free_pipeline(hcp);
      return;
    }
/*
 * If the parent got here then we the child process can run.  Set the child's
 * process info as running either in the foreground or the background (which
 * was done in parser.)  Also store child_pid number.
 */
    hcp->pid = child_pid;

/*
 * If the process is to run in the background, just do so and return
 * immediately.  The each_line() function will take care of checking and
 * updating the status of that process in the jobs table.
 */
    if (CUR_BG(hcp)) {
      fprintf(stderr, "[%d]\t %d\n", cur_job+1, child_pid);
      add_stat_to_proc(hcp, P_NOCHANGE, P_BG, 0); /* flag proc as running */
      was_bg = 1;
      next_fg = cur_job;
      return;
    }

/*
 * Now we have to wait for the child process to return.  However, it could
 * be that another backgrounded child has terminated and so we have to make
 * sure we actually wait for the process which is in the foreground now.
 */
    status = -2;
    while (status != child_pid) {
      status = wait3(&retval, WUNTRACED , NULL_RUSAGE);
      if (status <= 0)
	continue;
      tcp = get_head_proc(status);
				/* update status of processes */
/*      printf("EZK: %d %d %d\n",
	     tcp->cur_stat, tcp->new_stat, tcp->wait3_val);
*/
      if (WIFEXITED(retval)) {
	add_stat_to_proc(tcp, P_NOCHANGE, P_EXIT, retval);
      }
      if (WIFSTOPPED(retval)) {
	add_stat_to_proc(tcp, P_NOCHANGE, P_STOP, retval);
      }
      if (WIFSIGNALED(retval)) {
	add_stat_to_proc(tcp, P_NOCHANGE, P_SIG, retval);
      }
/*      printf("EZK: %d %d %d\n",
	     tcp->cur_stat, tcp->new_stat, tcp->wait3_val);
*/
#ifdef ESH_DEBUG
/*      fprintf(stderr, "PARENT(wait): status=%d\n", status);
      if (WIFEXITED(retval))
	fprintf(stderr, "PARENT(wait): PID %d exited with status %d\n",
		status, WEXITSTATUS(retval));
      if (WIFSTOPPED(retval))
	fprintf(stderr, "PARENT(wait): PID %d stopped by signal %d\n",
		status, WSTOPSIG(retval));
      if (WIFSIGNALED(retval))
	fprintf(stderr, "PARENT(wait): PID %d terminated by signal %d\n",
		status, WTERMSIG(retval));
*/
#endif /* ESH_DEBUG */
    }				/* end of while (status != child_pid)... */
    if (status < 0) {		/* wait ended due to an error */
      PERROR("PARENT: wait3 error");
      free_pipeline(tcp);
      return;
    }

/*
 * Get back control over terminal.
 */
    signal(SIGTTOU, SIG_IGN);
    status = tcsetpgrp(open_fd(), pid);	/* pid is the shell's pid */
    if (status < 0) {		/* cannot set terminal process group */
      PERROR("PARENT: tcsetpgrp failed");
      return;
    }
    
  } else {			/* fork succeeded.  We're in the CHILD. */
    pid = getpid();
    ppid = getppid();
    pgrp = getpgrp(pid);
    tcgrp = tcgetpgrp(open_fd());
#ifdef ESH_DEBUG
/*    fprintf(stderr, "CHILD(fork): PID=%d PPID=%d PGID=%d TCGRP=%d.\n",
	   pid, ppid, pgrp, tcgrp);
*/
#endif /* ESH_DEBUG */

/*
 * In order for the child to receive signals that do not affect the parent
 * shell, we must change the process group of the child first, before
 * execve-ing it.
 * For example, a SIGTERM (^C) should terminate the child process (if it has
 * not trapped that one) and the parent should be informed of that.
 */
    status = setpgrp(pid, pid);
    if (status < 0) {		/* cannot set the process group */
      PERROR("CHILD: setpgrp failed");
      exit(errno);
    }

/*
 * If the process is to run in the foreground, then the child should
 * take control over the tty.  We have to ignore some signal(s)
 * because otherwise the default action for tcsetpgrp is to send a
 * SIGTTOU which causes the syscall/ioctl inside tcsetpgrp to restart.
 * This gets you in an infinite loop.  That's why we have to ignore
 * the SIGTTOU signal.
 */
    if (CUR_FG(hcp)) {
      signal(SIGTTOU, SIG_IGN);
      status = tcsetpgrp(open_fd(), pid);
      if (status < 0) {		/* cannot set terminal process group */
	PERROR("CHILD: tcsetpgrp failed");
	exit(errno);
      }
    }

#ifdef ESH_DEBUG
/*    fprintf(stderr, "CHILD(execve): PID=%d PPID=%d PGID=%d TCGRP=%d.\n",
	    pid, ppid, pgrp, tcgrp);
*/
#endif /* ESH_DEBUG */
    signal(SIGINT, SIG_DFL);	/* child should NOT ignore these! */
    signal(SIGTSTP, SIG_DFL);
    status = execve(path, hcp->argv, sh_envp);
    if (status < 0) {		/* execve failed */
      PERROR("CHILD: execve failed");
      exit(errno);
    }
  }
}


/*
 * Check for processes that changed their status.
 */
static void status_check()
{ register int i;
  proc_t *tcp;

#ifdef ESH_DEBUG
/* print_pipeline(jobs[0].head_proc); */
/* print_pipeline(jobs[1].head_proc); */
/* print_pipeline(jobs[2].head_proc); */
#endif /* ESH_DEBUG */

  for (i=0; i<max_jobs; ++i) {
/* puts("IN FOR LOOP"); */
    tcp = jobs[i].head_proc;
    if (tcp == NULL_PROC)
      continue;

/* printf("EZK.4: %s %d %d %d\n", jobs[i].head_proc->argv[0],
       jobs[i].head_proc->cur_stat, jobs[i].head_proc->new_stat,
       jobs[i].head_proc->wait3_val); */

		/* check if process was running and is now stopped */
    if (CUR_RUN(tcp) && NEW_STOP(tcp)) {
/* puts("status_check sez process was running and is stopped now."); */
      print_proc_status(i, tcp);
    }
		/* check if process was running and was signaled to death */
    if (CUR_RUN(tcp) && NEW_SIG(tcp)) {
/* puts("status_check sez process was running and is DEAD now."); */
      print_proc_status(i, tcp);
      free_pipeline(tcp);		/* free process allocation */
      continue;
    }
		/* check if process was running and is now dead */
    if (CUR_RUN(tcp) && NEW_EXIT(tcp)) {
/* printf("Status_Check: \"%s\"was run and now exit!!!",
       tcp->argv[0]);  */
#ifdef NOTDEF
      if (CUR_BG(tcp))		/* don't print line if proc was in fg */
	print_proc_status(i, tcp);
#endif /* NOTDEF */
      free_pipeline(tcp);		/* free process allocation */
      continue;
    }
				/* copy old status to new one */
    tcp->cur_stat = tcp->new_stat;
  }
}


/*
 * Check if any other child processes have changed status and report it.
 */
void get_new_status()
{ int status, retval;
  register int i;
  proc_t *tcp;

  status = 0;
  do {
    status = wait3(&retval, WUNTRACED | WNOHANG , NULL_RUSAGE);
    if (status > 0) {		/* then a child process did change */
      tcp = get_head_proc(status);
				/* update status of processes */
/*      printf("EZK: %d %d %d\n",
	     tcp->cur_stat, tcp->new_stat, tcp->wait3_val); */
      if (WIFEXITED(retval)) {
	add_stat_to_proc(tcp, P_NOCHANGE, P_EXIT, retval);
      }
      if (WIFSTOPPED(retval)) {
	add_stat_to_proc(tcp, P_NOCHANGE, P_STOP, retval);
      }
      if (WIFSIGNALED(retval)) {
	add_stat_to_proc(tcp, P_NOCHANGE, P_SIG, retval);
      }
/*      printf("EZK: %d %d %d\n",
	     tcp->cur_stat, tcp->new_stat, tcp->wait3_val); */
#ifdef NOTDEF & ESH_DEBUG
      fprintf(stderr, "GET_NEW_STATUST(wait): status=%d\n", status);
      if (WIFEXITED(retval))
	fprintf(stderr, "GET_NEW_STATUS(wait): PID %d exited with status %d\n",
		status, WEXITSTATUS(retval));
      if (WIFSTOPPED(retval))
	fprintf(stderr, "GET_NEW_STATUS(wait): PID %d stopped by signal %d\n",
		status, WSTOPSIG(retval));
      if (WIFSIGNALED(retval))
	fprintf(stderr,
		"GET_NEW_STATUS(wait): PID %d terminated by signal %d\n",
		status, WTERMSIG(retval));
#endif /* ESH_DEBUG */
    }				/* and of if (status > 0) ... */
#ifdef NOTDEF
    if (status < 0) {		/* wait ended due to an error */
      PERROR("get_new_status: wait3 error");
      free_pipeline(tcp);
      return;
    }
#endif /* NOTDEF */
  } while (status > 0);

/*
 * Now cycle through job queue and free structures belonging to
 * processes which are dead.
 */
  for (i=0; i<max_jobs; ++i) {
    tcp = jobs[i].head_proc;
    if (tcp == NULL_PROC)
      continue;
    if (CUR_RUN(tcp) && NEW_EXIT(tcp)) {
      if (CUR_BG(tcp))		/* don't print line if proc was in fg */
	print_proc_status(i, tcp);
      free_pipeline(tcp);		/* free process allocation */
    }
  }
}


/*
 * Check on things that have to be done each time a processed line
 * terminates.  Checks if any processes changed their status, prints
 * prompt line, etc...
 */
void each_line()
{
/* puts("IN EACH_LINE before status_check"); */
  status_check();		/* check on the last process forked */
/* puts("IN EACH_LINE before get_new_status"); */
  get_new_status();		/* check for status on other processes */
/* puts("IN EACH_LINE before shell_prompt"); */
  shell_prompt();
/* puts("EXITING EACH_LINE"); */
}


/*
 * Find the job's head process given a pid.
 * Assume there must be one, or else exit.
 */
proc_t *get_head_proc(pid)
     int pid;
{ register int i;
  proc_t *tcp, *ncp;

  for (i=0; i<max_jobs; ++i) {
    tcp = jobs[i].head_proc;
    if (tcp == NULL_PROC)
      continue;
    ncp = tcp;
    while (ncp) {
      if (ncp->pid == pid)
	return ncp;
      ncp = ncp->next;
    }
  }
				/* had better found them by now */
  fprintf(stderr,
	  "cannot find job in table. job table inconsistency!\n");
  exit(-1);
}


/*
 * Find the job entry in the queue, given a pointer to the head proc.
 * Assume there must be one, or else exit.
 */
int get_job_num(hcp)
  proc_t *hcp;
{ register int i;
  proc_t *tcp;
  
  for (i=0; i<max_jobs; ++i) {
    tcp = jobs[i].head_proc;
    while (tcp) {		/* check the whole pipeline */
      if (tcp == hcp)
	return i;
      tcp = tcp->next;
    }      
  }
				/* had better found them by now */
  fprintf(stderr,
	  "cannot find job number in table. job table inconsistency!\n");
  exit(-1);
}
