/*
 * lock_client.c:
 *
 * Code for the client side of the external pager lock manager.
 * (based on "randon server" examples in /usr/mach/doc.)
 * Erez Zadok, Spring 1992.  e6118y.
 */

#include <mach/mach_param.h>
#include <mach/message.h>
#include <mach/notify.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <servers/netname.h>
#include <mach.h>
#include <mach_error.h>
#include <mig_errors.h>
#include <cthreads.h>
#include <stdio.h>
#include <libc.h>
#include <strings.h>

#include "xp_lock_manager.h"	/* decls of xp lock manager */
#include "locker.h"		/* generated by mig */
#include "xp.h"			/* decls of xp manager */

extern void print(), fprintf(), puts();

port_t xp_port = PORT_NULL;	/* memory object */
vm_size_t twopages;
static char *host;

void usage(s)
     char *s;
{ fprintf(stderr, "usage: %s [-h hostname]\n", s);
}

void panic(s)
     char *s;
{ fprintf(stdout, "panic: %s\n", s);
  exit(1);
}


/*
 * Perform the appropriate xp lock manager request.
 */
int xp_lock_mgr_request(request_code)
     int request_code;
{ static port_t xp_lock_mgr_port = PORT_NULL;
  kern_return_t retcode;
  int inask, outask;

				/* test proper call request */
  switch (request_code) {
  case XP_READ_LOCK:
  case XP_READ_UNLOCK:
  case XP_WRITE_LOCK:
  case XP_WRITE_UNLOCK:
    break;
  default:			/* this should not happen */
    printf("fatal error in decoding of extern pager lock request\n");
    exit(1);
  }
				/* find xp lock manager */
  if (xp_lock_mgr_port == PORT_NULL) {
    printf("Looking up port for xp lock manager\n");
    while (netname_look_up(name_server_port, host, XP_LOCK_MANAGER_NAME,
			   &xp_lock_mgr_port) != NETNAME_SUCCESS) {
      fprintf(stdout,"client: waiting for %s@%s to register\n",
	      XP_LOCK_MANAGER_NAME, host);
      sleep(10);
    }
    printf("xp_lock_mgr_port is %d\n", xp_lock_mgr_port);
  }

  inask = request_code;
  outask = 28;
  retcode = ask_lock_manager(xp_lock_mgr_port, inask, &outask);
  printf("from server return code %d; inask %d; outask is %d\n",
	 retcode, inask, outask);
  return retcode;
}


void main(argc, argv)
     int argc;
     char **argv;
{
  int request_code, i, c, errflg=0, test=1, starting_address, ending_address;
  char *p, cmd[80], hostname[MAXHOSTNAMELEN];
  boolean_t initialized=FALSE;
  kern_return_t retcode, err;
  vm_address_t data;
  extern int optind;
  extern char *optarg;

			/* by default, will look for server on local host */
  gethostname(hostname, MAXHOSTNAMELEN);
  host = hostname;

				/* get server's host and which test */
  while ((c = getopt(argc, (const char **) argv, "h:")) != EOF)
    switch (c) {
    case 'h':
      host = optarg;
      break;
    case '?':
      errflg++;
      break;
    }
  if (optind < argc)
    test = atoi(argv[optind]);
  if (errflg) {
    usage(argv[0]);
    exit(1);
  }
  printf("client: test %d host '%s'\n", test, host);

			/* get a port to the server from netname service */
  while (netname_look_up(name_server_port, host, XP_SERVER_NAME, &xp_port)
	 != NETNAME_SUCCESS) {
    fprintf(stdout,"client: waiting for %s@%s to register\n",
	    XP_SERVER_NAME, host);
    sleep(10);
  }

  twopages = 2*getpagesize();

				/* start command line interaction */
  while (!feof(stdin)) {
    printf("request? ");
    gets(&cmd[0]);
    if (feof(stdin)) {
      printf("exiting...\n");
      exit(1);
    }
    if (!strcmp(cmd, "+r"))
      request_code = XP_READ_LOCK;
    else if (!strcmp(cmd, "-r"))
      request_code = XP_READ_UNLOCK;
    else if (!strcmp(cmd, "+w"))
      request_code = XP_WRITE_LOCK;
    else if (!strcmp(cmd, "-w"))
      request_code = XP_WRITE_UNLOCK;
    else if (!strcmp(cmd, "read"))
      request_code = XP_READING;
    else if (!strcmp(cmd, "write"))
      request_code = XP_WRITING;
    else if (!strcmp(cmd, "q")) {
      request_code = XP_QUIT;
    } else
      request_code = XP_NO_CMD;


    switch (request_code) {

    case XP_READ_LOCK:
    case XP_WRITE_LOCK:
      if (!initialized) {	/* map the object */
	puts("mapping vm...");
	data = 0;
				/* causes MO_INIT() to be called */
	if ((err = vm_map(task_self(), &data, twopages, 0, TRUE, xp_port, 0,
			  FALSE, VM_PROT_ALL, VM_PROT_ALL, VM_INHERIT_SHARE))
	    != KERN_SUCCESS) {
	  mach_error("client: can't map the object - ", err);
	  exit(1);
	}
	fprintf(stdout, "client: object mapped at address 0x%lx\n", data);
	initialized = TRUE;
      }
				/* ask for un/lock */
      retcode = xp_lock_mgr_request(request_code);
      if (retcode == KERN_SUCCESS) {
	printf("un/lock request succeeded\n");
      }
      break;

    case XP_READ_UNLOCK:
    case XP_WRITE_UNLOCK:
/*
 * Deallocate vm and uninitialize.  This is so written changes get flushed
 * to the memory manager so that any subsequent reading will reflect the
 * changes.
 */
      if (initialized) {
	puts("deallocating vm...");
	err = vm_deallocate(task_self(), data, (vm_size_t) twopages);
	if (err != KERN_SUCCESS) {
	  mach_error("client: can't deallocate the object - ", err);
          exit(1);
	}
	initialized = FALSE;
      }
				/* ask for un/lock */
      retcode = xp_lock_mgr_request(request_code);
      if (retcode == KERN_SUCCESS) {
	printf("un/lock request succeeded\n");
      }
      break;

    case XP_READING:
      printf("\t read starting address: ");
      gets(&cmd[0]);
      starting_address = atoi(cmd);
      printf("\t read ending address: ");
      gets(&cmd[0]);
      ending_address = atoi(cmd);
      p = (char *) data;
      for (i=starting_address; i <= ending_address; i++) {
	printf("%c", p[i]);
      }
      printf("\n");
      break;

    case XP_WRITING:
      printf("\t write starting address: ");
      gets(&cmd[0]);
      starting_address = atoi(cmd);
      printf("\t write text: ");
      gets(&cmd[0]);
      p = (char *) data;
      strncpy(&p[starting_address], &cmd[0], strlen((const char *) &cmd[0]));
      break;

    case XP_QUIT:
      puts("quiting...");
      exit(1);
      break;			/* not needed */

    case XP_NO_CMD:
    default:
      printf("\tCommands are [+r | +w | -r | -w | read | write | q]\n");
      break;
    }				/* end of switch (request_code) */
  }				/* end of while (!feof(stdin)) */
  exit(0);			/* in case you hit ^D */
}
