@device [postscript]
@make [article]
@style [spacing 1.5, indent 5, fontfamily timesroman]
@style [fontscale=10]
@blankspace [3 lines]
@heading [Advanced Operating Systems E6118y]
@blankspace [2 lines]
@majorheading [Assignment 4]
@majorheading [A criticism of Mach's EMMI]
@blankspace [3 line]
@heading [@+{by}]
@heading [Erez Zadok (Hebrew Wank 1)]
@blankspace [1 line]
@heading [Due: March 10, 1992]
@blankspace [3 lines]

@NewPage

During the last assignment, I learned how to use Mach's EMMI (External
Memory Manager Interface.)  This is a criticism of its advantages and
disadvantages.

@Section[The External Memory Manger Interface]

@SubSection[Hard to understand and debug]

The interface is hard to understand.  It very much reminds me the same
problems I had when I first used RPC or YACC.  Many things are hidden from
you the user as you program in a ``high-level'' language.  A lot of
intermediate steps happen that you may not be aware of.  Unless you are
already an expert with that interface, you may find it very hard to
understand @B[exactly] what is happening at each step without tracing all of
the generated code line by line.  For example, when I have a problem with my
YACC code, I often resort to tracing the C code it generates in order to
understand what it is doing.

The problem is even harder with the EMMI.  There is really no intermediate C
code generated for you that you can examine.  You have to resort to kernel
debugging methods, namely @T[printf()]'s...  Furthermore, the kernel is the
main initiator of the EMM's functions, and being asynchronous just makes
chronological debugging harder.

It took us (``@I[The Fab Five]'') almost five days to read all the
information and trace all the examples until we thoroughly understood every
aspect of the EMMI's behavior.  At that point we also realized that some
things simply cannot be done with that interface.  It seems to me that any
kind of system ought to be easily understood by one person in one day--not
an order of magnitude longer.

@SubSection[Lack of information]

The man pages are in a sorry state.  In the process of creating or
formatting them, arbitrary (or so it seems) sections of text were omitted.
I even FTP-ed the original man pages from the net and found them to be
erroneous as well.@Foot[I would hope they were fixed in 3.0.]  The only
sources of accurate information were a few Postscript files I FTP-ed from
the net and a few more placed in the class library.

the documentation that is good and available is still not sufficient.  It is
very simple, lacks in examples, and assumes that you are already familiar
with Mach internals.  As a first time user of Mach, I had to abandon my
previous knowledge of Unix ``ports'', ``clean'' and ``flush''--Mach uses
them in a different way.

I FTP-ed a few more examples of external pagers written with the EMMI.
However, they were all virtually identical, managing a single predefined
object and not having any locking protocol.  In fact, in the README file of
one of the examples, the author was complaining about the same lack of
information from the CMU people.

@SubSection[Locking and security]

My main complaint about this interface is the lack of built-in locking
mechanism and security considerations within the design.  The EMM only
communicates with its local kernel.  The kernel passes no information that
can identify the calling client:

@Begin[Itemize]

There is no way to find out the port of the calling client.  This means that
the EMM cannot store a list of privileged clients and selectively allow or
deny them access.

When the EMM gets a request, it may grant it or not, but whatever decision
it makes will affect the whole mapped object for @B[all] clients who mapped
that object.  For example, even if the EMM had a way of knowing which client
called it, if it then decided to grant only read permission to the client,
it would have affected everyone who would try to write the object later.
There is no way to restrict access permissions per client.

There is no way to tell if the calling client is on the same host, or on a
remote host (and if so, which remote host it is.)  The only way that can
partially be found is when a first time client on a remote host starts up,
it causes @T[mo_init()]@Foot[I use the notation @T{mo_}@I{function}@T{()} to
indicate the full-name function @T{memory_object_}@I{function}@T{()}.] to
occur in order for the actual data of the shared object to be transferred to
the remote kernel.  Still no indication is available to tell the EMM who
called it--it only knows that a new kernel has asked to map the object (the
new kernel possibly being the local one.)  However, the kernel calls the EMM
with a different port per calling host, so access persmissions could be
based on whole hosts (but not on tasks within.)

Access to the object is given on a per-page basis.  Even if all you asked is
to read or write the first byte of the page, you get the whole page.  It
seems that more in line with database systems, access should be for the
particular byte-range you asked and no more (i.e., locking a record or
field.)  This however, can be implemented using available resources, given
an extra level of indirection.

One of the more non-obvious implications of this interface is the ability to
ask for one set of permissions to access an object, to which the EMM may
reply with a different set of permissions, which may even be orthogonal to
the requested set.  That is, you may ask for read-only, but receive
read/write, write-only, or no permissions at all, @B[even] though the object
was successfully mapped into your address space.  This triggers one of the
more serious bugs in the implementation of the interface.  For example, if
you asked for write permission but was only given read-permission, and then
you try to write to the object, your client process is hung.  At that point,
the kernel on your behalf will issue a @T[mo_data_request()] followed by a
@T[mo_data_unlock()] to the EMM.  If the EMM will still not grant write
access to the object, the client will still stay hung, and the kernel will
reissue that request, and so on ad infinitum.  That infinite loop takes up
lots of resources.  Furthermore, there has to be a way for the EMM to tell
the kernel that to stop asking for the permission, to delay the asking for
some time, or to tell the kernel to send some signal to the client (which is
hung) to notify it that the object is not available and probably won't be.

@End[Itemize]

Given all of the above, there is no way to perform intelligent and
client-independent locking on the object being managed.  A different locking
protocol had to be built (which I chose to implement as a separate task.)
All clients had to know of the protocol and follow it; they had to tell the
lock-manager when they wanted shared or exclusive lock on the object being
managed.  They had to release it in time so that other clients could access
the object as well.  There was no way to always catch and avoid malicious
clients which did not use my locking protocol.

@Paragraph[MIG]

I found out MIG to be easy to use.  It did what was required of it and
worked well.  However, for the creative programmer, supplying only one level
of message passing abstraction is insufficient.  In my opinion, Sun's RPC
solved this problem elegantly by providing the programmer with three levels
of abstraction.  Each level offers a different combination of simplicity vs.
controllability.  For example, in the highest level of RPC (using
@T[rpcgen]) you only need to specify a few lines of code and the rest of the
client and server code will be generated for you.  However, you cannot
change the transport protocol used, timeouts, or retransmissions.  For that
you need to go one or two levels down.  But if you do, you will have to
write more code yourself.

Even so, I still found MIG quite useful.  The hardest part about it was the
need to write my own @T[mig_server()] function.  This one is supposed to
perform the receive-process-send message loop.  I could not use the default
@T[mig_server()] function because of the locking protocol.  I had to accept
all locking requests, queue them, and then process some or none of them in
order, essentially decoupling the receiving and sending of messages.  Some
trickery was needed in order to save the whole context of the just received
message so that it can be reprocessed later as if it just arrived.  At the
cost of a bit more complicated @T[mig_server()] code, I achieved one of my
goals which was not to have to modify the MIG-generated code.

@SubSection[No easy way to manage multiple objects]

The EMMI does not support management of multiple objects at the same time.
The clients have no idea what object they asked to map; all the clients know
is the name of the EMM as registered with the name-server.  That name can
be, but is not necessarily dependent on the object being managed.

If one were to manage multiple objects (say, for simplicity, files) then a
different messaging interface would have had to be built into my EMM which
allows a client to ask for a particular file to be managed, list and choose
from all available files being managed at any point, and finally tell the
EMM to terminate the management of a file object.  I was really surprised to
find out it was not built into the EMMI.  In my opinion this is one of the
first things any EMMI user would like to have.

@SubSection[EMMI's options to functions]

I found out that the division of configuration options in the EMMI is
confusing and sometimes illogical.  It seems that some options are in the
wrong place, while others may be specified in several places:

@Begin[Itemize]

The man pages for @T[mo_object_init()] tell you that you should call
@T[mo_set_attributes()] if you want the kernel to process any requests on
this object.  Every implementation of this initialization routine also sets
up the proper attributes for the object.  If that is the case, why make
these two separate functions?  I think that any feature that is rarely (if
at all) used was not properly designed.

The @T[may_cache_object] option to @T[mo_set_attributes()] should be more
carefully explained.  If used to manage an object over several remote hosts,
and this option is @T[TRUE], each remote host may keep its own version of
the object once it is written to locally.  The main version on the EMM
machine will not be ``in sync'' with it.

Some of the cache management functions to the EMMI were given to the
@T[mo_lock_request()] function via the @T[should_clean] and @T[should_flush]
options, but another cache control was placed in the @T[may_cache_object]
option to @T[mo_set_attributes()] function.  These cache controls should
have been consolidated into one EMMI call.  Furthermore, among this three
flags, some combinations simply make no sense, although they do not generate
errors.  For example, if you turn off @T[may_cache_object], then
@T[should_flush] makes no sense whether @T[TRUE] or @T[FALSE].

@T[Mo_data_request()] asks for some @T[desired_access] which is a positive
value.  The EMM may decide to retrieve the data, but give it a different
permissions (I already talked about that.)  The problem here is that when
@T[mo_data_provided()] is used to tell the kernel which permissions were
granted to the object in the out-parameter @T[lock_value], they are
@B[negative] options.  That is, that same bit that was used to ask for some
permission has to be negated when telling the kernel which options were
granted.  This opens up the door to lots of confusion and trivial but hard
to find bugs.  In my opinion, if a client asked for a particular set of
permissions, it should either get it or none at all.  A reply back to the
kernel should not also specify a change in access to the object.  That
functionality is the task of @T[mo_data_unlock()] and @T[mo_lock_request()].

The same permission confusion problem exists in @T[mo_data_unlock()] and
@T[mo_lock_request()]--the first specifies positive permissions and the
latter negative ones.  These permission changing functions unnecessarily
exist in @T[mo_data_request()] and @T[mo_data_provided()].  However, these
data un/lock functions are the right place for permission requests.  My
complaint is similar as before; a permissions should be granted exactly as
requested or none at all.  The out-parameter @T[lock_value] is not needed.

@End[Itemize]

@SubSection[Client-level interface]

I've talked about the need for the clients to be cooperative with respect to
object locking and access.    There are a few more complaints:

@Begin[Itemize]

It is quite silly for @T[vm_map()] to be able to ask for a set of
permissions which may not be granted at all, while at the same time the
mapping is successful.  The moment you try to access the object you may be
indefinitely blocked.  In my opinion the @T[vm_map()] call should never
return if the right permissions could not have been granted.

No matter what caching strategy the EMM chose, the client cannot tell the
EMM to flush any changes back to permanent media (particularly useful in the
case of file objects.)  If the EMM defined @T[may_cache_object] as @T[TRUE]
when setting the attributes, then nothing the client will do (even
@T[exit()]) will cause the kernel to tell the EMM to write the data back to
disk.  If @T[may_cache_object] is set to @T[FALSE] then only when the client
terminates or explicitly calls @T[vm_deallocate()] that the kernel tells the
EMM to write the modified data back to disk.  But then if the client still
wants to use the mapped object, it has to call @T[vm_map()] again.  This
need is expensive and unnecessary in my opinion.  A function such as
``@T[vm_flush()]'' should have been provided to allow the synchronization of
the cached data with the permanent media without the need for unmapping of
the object.  In fact in my implementation I had to set the the options
carefully in order to make sure that all kernels and the copy on disk of the
file object being managed are in as much in sync with each other as
possible.

@End[Itemize]

@SubSection[Synchronous vs. asynchronous interfaces]

The client-to-kernel interface is synchronous.  The kernel-to-EMM one is
asynchronous.  There is no way to change that.  I believe in flexibility and
think there should have been a way to specify what kind of communication you
want to have.

A smart client might want asynchronous communication with the kernel.  It
way elect to post requests for object access which may be granted later,
then go on and perform some other operation.  When the kernel was ready to
grant the particular access, it could have notified the client via some
signal that the access was granted.  This would especially prove useful in
an environment where data-locking is provided--the client would not have to
hang while waiting (possibly indefinitely) for access permission.

I cannot see any real need for allowing the kernel-to-EMM interface to be
synchronous.  If the EMM has to wait for the kernel to respond back, it may
not be able to service other clients.  If the kernel has to wait for the EMM
to respond, it may hang forever (what if the EMM crashed?)  The last thing
anyone wants is for the deadlock in the kernel.

@Section[But it works!]

@SubSection[Ease of use vs. non-kernelized operating systems]

The EMMI actually works.  You can get quite a lot of it, even with its
limitations.  Starting your EMM and activating it is smooth.  The interface
gives you a set of primitive calls of which you can combine to perform more
complicated operations.

It is a lot easier to test out new external paging schemes using this
interface.  In fact, it seems to me to be the primary reason CMU came up
with it.  User-level debugging is an order of magnitude easier as compared
with in-kernel interfaces (speaking from personal experience...)

Furthermore, it is even possible to have different memory management schemes
for different objects.  With the many options available in the EMMI, one may
tailor one's particular EMM depending on the object(s) being managed.  For
example, for high memory-to-disk consistency, one may configure the EMM to
always flush changes back to disk.  For critical read-only objects one may
only give out read permissions.  If the EMM is to have as much control and
knowledge of what the kernel is doing with the object(s), one can configure
the EMM to have the kernel notify it when a lock request has completed, or
even write back any modified pages back to the EMM.  Certainly enough
flexibility was given out to the EMM implementor.

@Section[Other possible interfaces]

@SubSection[Practicality]

The whole idea of pulling major interfaces out of the kernel is novel.  It
makes for easier development and testing of any of the pieces that make up
the operating system.  Although kernelized operating systems are still
younger than non-kernelized ones, and not as fast or optimized, they have a
few distinct advantages.  Kernelized systems can be out of the debugging
stage and into the production market in shorter time.  The users of these
systems need not be inconvenienced every time a new kernel patch is posted
or when a new version of a kernel service is available.  Users need not know
or detect any system disruption while the system administrator is installing
a new version of the EMM or AFS.

@SubSection[How much can be taken out?]

I would first argue that reasons of performance should not stop one from
taking a kernel service outside.@Foot[unless you are talking about vast
performance difference... your mileage may vary.]  Memory is getting
cheaper, CPUs faster, and disks larger every day.  Unacceptable performance
today may be more than acceptable a year from now, especially taking into
account the fact that kernelized systems may be less buggy and more stable
in the long run, being easier to maintain.

Certain services are already available outside the kernel,@Foot[at least in
Mach 3.0] such as Networking, Filesystems (NFS, AFS, etc), External Memory
Management, and even simulation of other operating systems.  Other services
may never venture outside the kernel, by their design nature.  Such include
Booting and Autoconfiguration.

@SubSection[An ``autoloader'' service]

But what about other services such as the scheduler, devices driver, or
Interprocess Communication?  I argue that at the very least there will have
to be a way to load up new modules into the kernel that will overwrite the
old implementation of these services with a new one.  Such an ``autoloader''
(which may be autoloadable itself!) will have to make sure that the system's
transformation from the old to the new implementation of the service is
smooth enough.  Actually, such autoloading need not clobber the old
implementation--it may keep it so one can revert to a known working service
in case of problems.

There are some problems with this kind of autoloading.  First, if anything
goes wrong it may result in loss of data.  These are certainly not the kind
of tests you want to perform on a production environment.  Then, it is
nothing but a smooth replacement of some portion of kernel code with another
(with the possibility of keeping the old one intact.)  It is not really an
out-of-kernel interface, but is resembles one, and is the next best thing.

@SubSection[Real out-of-kernel interfaces]

Whether an ``autoloader'' or out-of-kernel interface, one thing is clear to
me.  From the points of debuggability, security, and minimization of
disruptions, all newly tested interfaces should allow for the restriction of
their control.  Just as in the case of the EMMI, we can manage one object
alone, and get the bugs out of our system; the rest of the machine still
uses the default memory manager and sees no interruptions.  If we were to
write a new default memory manager, we would certainly first test it as a
specialized one managing a few objects, before we ``unleash' it on the rest
of the users.

@Paragraph[Out-of-kernel scheduling]

The BSD people who originally wrote Unix had a specific set of algorithms
for deciding which process to schedule first.  Some of their formulas and
algorithms make no sense to anyone today.  Perhaps at the time there were
good reasons, but not being able to change these conveniently is a problem.
A lot of optimization issues boil down to environment-specific tweaks.  For
example, a machine which is supposed to be a single-user's station running
X-windows most of the time need to give higher priority to that user's xterm
and emacs buffers.  But a large file-server aught to give higher priority
towards, I/O-, network-, and NFS-bound operations.  An email and NetNews
server should favor sendmail processes and those owned by user @T[news], and
so on.

You may argue that such scheduling decisions are controllable right now
using ``nice'' values.  That is partially true.  ``Nice'' values do not
translate to an obvious scheduling algorithm--they are only part of a more
complicated formula used by the scheduler.  You as a user-level process have
no control over that formula and other scheduling aspects.  Furthermore, if
you were to modify the kernel and code your own scheduler, it would still be
a static one.

An out-of-kernel scheduler first has to have the ability to control the
scheduling of a subset of processes (perhaps based on process groups or
``sessions''.)  Every process will be marked whether it is scheduled by the
default scheduler, or by another one (that we are presumably testing.)  The
first problem would be for the kernel to decide which scheduler's
process-subset to deal with at this point.  A simple part of the user-level
interface can be to determine the number of clock ticks allocated for any
particular scheduler.  For example, the user may specify that the default
scheduler will handle 9 ticks, and the 10@+[@U{th}] one will use your
private scheduler on the processes it is in control of only.  This simple
way will guarantee that 1 in every 10 ticks is yours.

When the time is up for your scheduler to decide on which process to
schedule next, the kernel will hand it a list of processes that are ready to
run, with all the information relevant to scheduling decisions (probably the
@T[proc] structure.)  Your interface will have to make a decision based on
the information and reply back to the kernel with the pid of the next
process to schedule.  This operation had better be fast and finish well
before the next clock tick.  The kernel will then run that process.

As a matter of optimization, more suitable for multiprocessor machines, the
kernel may even tell your scheduler of sleeping processes.  That way your
scheduler can make some early scheduling decisions when these processes will
be ready to run, and save time later.

With this interface, it seems possible to run several schedulers on the same
system at the same time.  Some may be pure test schedulers.  Others may
redefine your system (I.e., file-server-scheduler, single-user-scheduler,
mail-gateway-scheduler, netnews-scheduler, etc.)

@Paragraph[Out-of-kernel Interprocess Communication]

This one seems harder to take out of the kernel.  The a user-level interface
would have to have access and control over all message and posted signals
queues and be able to deliver these messages directly as much as possible to
other processes.  In fact, I don't see any specific advantages to taking
this out of the kernel, unless @B[direct] process communication can happen
without the kernel's intervention.  Unlike the case of a scheduler, there
aren't too many algorithms or optimization decisions that give a user-level
IPC an advantage over a kernel-based one.

@Paragraph[Out-of-kernel device-drivers]

This one of the cases where a ``backup'' or ``default'' device driver will
not do.  Access to I/O devices must be unique and centralized to maintain
system sanity.  If an out-of-kernel interface existed for such a service,
and you were allowed to replace, say, the SCSI device driver with your
own--the older one would have to cease to be used and all control would have
to be passed to the new driver.

Creating such an interface in a portable way will be difficult, given the
non-portable nature of device-drivers.  I think performance would degrade
considerably.  Furthermore, device-drivers tend to be written once and
sparingly modified.  I think this is one of the cases where ``autoloading''
will prove a better scheme than out-of-kernel interface.
