/* Copyright (C) 1988, 1989, 1992 Free Software Foundation, Inc.
This file is part of GNU Make.

GNU Make is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU Make is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Make; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "make.h"
#include "commands.h"

#include "/import/pkg/pvm3.1/include/pvm3.h"

static char rcsid[]=
"$Id: remote-pvm.c,v 1.5 1993/04/30 08:25:37 lih Exp lih $";

char *remote_description = "pvm";

#define MAXHOSTS    256
#define MAKEDAEMON  "maked"

/* Normalized path name of the current directory.  */
static char *normalized_cwd;

static int              nhost;
static struct hostinfo *hostp;	/* Points to static table */

static int bogus_id=666;

/* Return nonzero if the next job should be done remotely.  */

int
start_remote_job_p ()
{
  static int inited = 0;

  if (getenv ("GNU_MAKE_USE_PVM") == 0) {
    fprintf(stderr,"make: start_remote_job_p, using vanilla local jobs\n");
    return 0;
  }

  if (!inited) {
    
    int mytid, status, narch;
    
    hostp = (struct hostinfo *) malloc (sizeof (struct hostinfo) * MAXHOSTS);
    
    /* PVM Startup */
    mytid = pvm_mytid();
    
    status = pvm_config(&nhost, &narch, &hostp);
    
    if (status) {
      free (hostp);		/* Free memory */
      perror("pvm_config");
      return 0;
    }

    normalized_cwd = (char *) xmalloc (GET_PATH_MAX);
    strcpy (normalized_cwd, starting_directory);

    fprintf(stderr,"make: start_remote_job_p, using pvm!\n");
    inited++;
  }
  
  return 1;
}

/* Start a remote job running the command in ARGV,
   with environment from ENVP.  It gets standard input from STDIN_FD.  On
   failure, return nonzero.  On success, return zero, and set *USED_STDIN
   to nonzero if it will actually use STDIN_FD, zero if not, set *ID_PTR to
   a unique identification, and set *IS_REMOTE to zero if the job is local,
   nonzero if it is remote (meaning *ID_PTR is a process ID).  */

int
start_remote_job (argv, envp, stdin_fd, is_remote, id_ptr, used_stdin)
     char **argv, **envp;
     int stdin_fd;
     int *is_remote;
     int *id_ptr;
     int *used_stdin;
{
  int  sbuf;			/* PVM send buffer */
  int  i, n;			/* Loop control variables */

  fprintf(stderr,"start_remote_job: *** ENTERED\n");

#ifdef FUZ
  if (!stdin_fd) {
    fprintf(stderr,"start_remote_job: stdin_fd = 0, not doing anything\n");
    return -1;
  }

  sbuf = pvm_mkbuf(PvmDataDefault); /* Create a Default encoding buf */
  (void) pvm_setsbuf(sbuf);         /* Set to be current buffer      */

  /* Pack working directory
   */
  pvm_pkstr(normalized_cwd);

  /* Pack argv
   */
  if (argv)
    for (n = 0; argv[n]; n++);
  else
    n = 0;
  pvm_pkint(&n, 1, 1);		         /* Pack the length of vector */
  for (i = 0; i < n; i++)	         /* Pack each string */
    pvm_pkstr(argv[i]);

  /* Pack envp
   */
  if (envp)
    for (n = 0; envp[n]; n++);
  else
    n = 0;
  pvm_pkint(&n, 1, 1);		         /* Pack the length of vector */
  for (i = 0; i < n; i++)	         /* Pack each string */
    pvm_pkstr(envp[i]);

  pvm_spawn
#endif

  fprintf(stderr,"start_remote_job: *** LEAVING\n");

  /*
   */
  *is_remote=1;
  {
    *id_ptr=bogus_id;
    bogus_id++;
  }
  *used_stdin=0;

  /* return -1; */
  return 0;
}

/* Get the status of a dead remote child.  Block waiting for one to die
   if BLOCK is nonzero.  Set *EXIT_CODE_PTR to the exit status, *SIGNAL_PTR
   to the termination signal or zero if it exited normally, and *COREDUMP_PTR
   nonzero if it dumped core.  Return the ID of the child that died,
   0 if we would have to block and !BLOCK, or < 0 if there were none.  */

int
remote_status (exit_code_ptr, signal_ptr, coredump_ptr, block)
     int *exit_code_ptr, *signal_ptr, *coredump_ptr;
     int block;
{
  fprintf(stderr,"remote_status: called with block=%d\n", block);

  if (block)
    fprintf(stderr,"remote_status: make wants us to block\n");
  
  *exit_code_ptr=0;
  *signal_ptr=0;
  *coredump_ptr=0;

  /* return -1; */
  if (bogus_id==666)
    return 0;
  else
    return --bogus_id;
}

/* Block asynchronous notification of remote child death.
   If this notification is done by raising the child termination
   signal, do not block that signal.  */
void
block_remote_children ()
{
  return;
}

/* Restore asynchronous notification of remote child death.
   If this is done by raising the child termination signal,
   do not unblock that signal.  */
void
unblock_remote_children ()
{
  return;
}

/* Send signal SIG to child ID.  Return 0 if successful, -1 if not.  */
int
remote_kill (id, sig)
     int id;
     int sig;
{
  return -1;
}
